/**
 * Copyright 2019-2023 by XGBoost contributors
 */
#ifndef XGBOOST_VERSION_CONFIG_H_
#define XGBOOST_VERSION_CONFIG_H_

#define XGBOOST_VER_MAJOR 3  /* NOLINT */
#define XGBOOST_VER_MINOR 2  /* NOLINT */
#define XGBOOST_VER_PATCH 1  /* NOLINT */

#endif  // XGBOOST_VERSION_CONFIG_H_
